@@ -12,7 +12,7 @@
 #include "stdafx.h"
 #include "Sndfile.h"
 #include "plugins/PluginManager.h"
-#include "../common/StringFixer.h"
+#include "../common/mptStringBuffer.h"
 #include "../common/version.h"
 
 
@@ -23,7 +23,6 @@ struct UpgradePatternData
 {
 	UpgradePatternData(CSoundFile &sf)
 		: sndFile(sf)
-		, chn(0)
 		, compatPlay(sf.m_playBehaviour[MSF_COMPATIBLE_PLAY]) { }
 
 	void operator() (ModCommand &m)
@@ -46,7 +45,7 @@ struct UpgradePatternData
 		{
 			// Out-of-range global volume commands should be ignored in S3M. Fixed in OpenMPT 1.19 (r831).
 			// So for tracks made with older versions of OpenMPT, we limit invalid global volume commands.
-			if(version < MAKE_VERSION_NUMERIC(1, 19, 00, 00) && m.command == CMD_GLOBALVOLUME)
+			if(version < MPT_V("1.19.00.00") && m.command == CMD_GLOBALVOLUME)
 			{
 				LimitMax(m.param, ModCommand::PARAM(64));
 			}
@@ -54,8 +53,8 @@ struct UpgradePatternData
 
 		else if(modType & (MOD_TYPE_IT | MOD_TYPE_MPT))
 		{
-			if(version < MAKE_VERSION_NUMERIC(1, 17, 03, 02) ||
-				(!compatPlay && version < MAKE_VERSION_NUMERIC(1, 20, 00, 00)))
+			if(version < MPT_V("1.17.03.02") ||
+				(!compatPlay && version < MPT_V("1.20.00.00")))
 			{
 				if(m.command == CMD_GLOBALVOLUME)
 				{
@@ -83,14 +82,14 @@ struct UpgradePatternData
 			// In the IT format, slide commands with both nibbles set should be ignored.
 			// For note volume slides, OpenMPT 1.18 fixes this in compatible mode, OpenMPT 1.20 fixes this in normal mode as well.
 			const bool noteVolSlide =
-				(version < MAKE_VERSION_NUMERIC(1, 18, 00, 00) ||
-				(!compatPlay && version < MAKE_VERSION_NUMERIC(1, 20, 00, 00)))
+				(version < MPT_V("1.18.00.00") ||
+				(!compatPlay && version < MPT_V("1.20.00.00")))
 				&&
 				(m.command == CMD_VOLUMESLIDE || m.command == CMD_VIBRATOVOL || m.command == CMD_TONEPORTAVOL || m.command == CMD_PANNINGSLIDE);
 
 			// OpenMPT 1.20 also fixes this for global volume and channel volume slides.
 			const bool chanVolSlide =
-				(version < MAKE_VERSION_NUMERIC(1, 20, 00, 00))
+				(version < MPT_V("1.20.00.00"))
 				&&
 				(m.command == CMD_GLOBALVOLSLIDE || m.command == CMD_CHANNELVOLSLIDE);
 
@@ -105,8 +104,8 @@ struct UpgradePatternData
 				}
 			}
 
-			if(version < MAKE_VERSION_NUMERIC(1, 22, 01, 04)
-				&& version != MAKE_VERSION_NUMERIC(1, 22, 00, 00))	// Ignore compatibility export
+			if(version < MPT_V("1.22.01.04")
+				&& version != MPT_V("1.22.00.00"))	// Ignore compatibility export
 			{
 				// OpenMPT 1.22.01.04 fixes illegal (out of range) instrument numbers; they should do nothing. In previous versions, they stopped the playing sample.
 				if(sndFile.GetNumInstruments() && m.instr > sndFile.GetNumInstruments() && !compatPlay)
@@ -117,7 +116,7 @@ struct UpgradePatternData
 			}
 
 			// Command I11 accidentally behaved the same as command I00 with compatible IT tremor and old effects disabled
-			if(m.command == CMD_TREMOR && m.param == 0x11 && version < MAKE_VERSION_NUMERIC(1, 29, 12, 02) && sndFile.m_playBehaviour[kITTremor] && !sndFile.m_SongFlags[SONG_ITOLDEFFECTS])
+			if(m.command == CMD_TREMOR && m.param == 0x11 && version < MPT_V("1.29.12.02") && sndFile.m_playBehaviour[kITTremor] && !sndFile.m_SongFlags[SONG_ITOLDEFFECTS])
 			{
 				m.param = 0;
 			}
@@ -127,16 +126,16 @@ struct UpgradePatternData
 		{
 			// Something made be believe that out-of-range global volume commands are ignored in XM
 			// just like they are ignored in IT, but apparently they are not. Aaaaaargh!
-			if(((version >= MAKE_VERSION_NUMERIC(1, 17, 03, 02) && compatPlay) || (version >= MAKE_VERSION_NUMERIC(1, 20, 00, 00)))
-				&& version < MAKE_VERSION_NUMERIC(1, 24, 02, 02)
+			if(((version >= MPT_V("1.17.03.02") && compatPlay) || (version >= MPT_V("1.20.00.00")))
+				&& version < MPT_V("1.24.02.02")
 				&& m.command == CMD_GLOBALVOLUME
 				&& m.param > 64)
 			{
 				m.command = CMD_NONE;
 			}
 
-			if(version < MAKE_VERSION_NUMERIC(1, 19, 00, 00)
-				|| (!compatPlay && version < MAKE_VERSION_NUMERIC(1, 20, 00, 00)))
+			if(version < MPT_V("1.19.00.00")
+				|| (!compatPlay && version < MPT_V("1.20.00.00")))
 			{
 				if(m.command == CMD_OFFSET && m.volcmd == VOLCMD_TONEPORTAMENTO)
 				{
@@ -146,7 +145,7 @@ struct UpgradePatternData
 				}
 			}
 
-			if(version < MAKE_VERSION_NUMERIC(1, 20, 01, 10)
+			if(version < MPT_V("1.20.01.10")
 				&& m.volcmd == VOLCMD_TONEPORTAMENTO && m.command == CMD_TONEPORTAMENTO
 				&& (m.vol != 0 || compatPlay) && m.param != 0)
 			{
@@ -157,7 +156,7 @@ struct UpgradePatternData
 				m.param = mpt::saturate_cast<ModCommand::PARAM>(param);
 			}
 
-			if(version < MAKE_VERSION_NUMERIC(1, 22, 07, 09)
+			if(version < MPT_V("1.22.07.09")
 				&& m.command == CMD_SPEED && m.param == 0)
 			{
 				// OpenMPT can emulate FT2's F00 behaviour now.
@@ -165,7 +164,7 @@ struct UpgradePatternData
 			}
 		}
 
-		if(version < MAKE_VERSION_NUMERIC(1, 20, 00, 00))
+		if(version < MPT_V("1.20.00.00"))
 		{
 			// Pattern Delay fixes
 
@@ -173,7 +172,7 @@ struct UpgradePatternData
 			// We also fix X6x commands in hacked XM files, since they are treated identically to the S6x command in IT/S3M files.
 			// We don't treat them in files made with OpenMPT 1.18+ that have compatible play enabled, though, since they are ignored there anyway.
 			const bool fixX6x = (m.command == CMD_XFINEPORTAUPDOWN && (m.param & 0xF0) == 0x60
-				&& (!(compatPlay && modType == MOD_TYPE_XM) || version < MAKE_VERSION_NUMERIC(1, 18, 00, 00)));
+				&& (!(compatPlay && modType == MOD_TYPE_XM) || version < MPT_V("1.18.00.00")));
 
 			if(fixS6x || fixX6x)
 			{
@@ -204,8 +203,8 @@ struct UpgradePatternData
 		}
 
 		if(m.volcmd == VOLCMD_VIBRATODEPTH
-			&& version < MAKE_VERSION_NUMERIC(1, 27, 00, 37)
-			&& version != MAKE_VERSION_NUMERIC(1, 27, 00, 00))
+			&& version < MPT_V("1.27.00.37")
+			&& version != MPT_V("1.27.00.00"))
 		{
 			// Fix handling of double vibrato commands - previously only one of them was applied at a time
 			if(m.command == CMD_VIBRATOVOL && m.vol > 0)
@@ -230,17 +229,25 @@ struct UpgradePatternData
 			m.volcmd = VOLCMD_NONE;
 		}
 
+		// Previously CMD_OFFSET simply overrode VOLCMD_OFFSET, now they work together as a combined command
+		if(m.volcmd == VOLCMD_OFFSET && m.command == CMD_OFFSET && version < MPT_V("1.30.00.14"))
+		{
+			if(m.param != 0 || m.vol == 0)
+				m.volcmd = VOLCMD_NONE;
+			else
+				m.command = CMD_NONE;
+		}
 	}
 
 	const CSoundFile &sndFile;
-	CHANNELINDEX chn;
+	CHANNELINDEX chn = 0;
 	const bool compatPlay;
 };
 
 
 void CSoundFile::UpgradeModule()
 {
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 17, 02, 46) && m_dwLastSavedWithVersion != MAKE_VERSION_NUMERIC(1, 17, 00, 00))
+	if(m_dwLastSavedWithVersion < MPT_V("1.17.02.46") && m_dwLastSavedWithVersion != MPT_V("1.17.00.00"))
 	{
 		// Compatible playback mode didn't exist in earlier versions, so definitely disable it.
 		m_playBehaviour.reset(MSF_COMPATIBLE_PLAY);
@@ -249,22 +256,24 @@ void CSoundFile::UpgradeModule()
 	const bool compatModeIT = m_playBehaviour[MSF_COMPATIBLE_PLAY] && (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT));
 	const bool compatModeXM = m_playBehaviour[MSF_COMPATIBLE_PLAY] && GetType() == MOD_TYPE_XM;
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 20, 00, 00))
+	if(m_dwLastSavedWithVersion < MPT_V("1.20.00.00"))
 	{
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if(Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
 			ModInstrument *ins = Instruments[i];
+			if(!ins)
+				continue;
 			// Previously, volume swing values ranged from 0 to 64. They should reach from 0 to 100 instead.
-			ins->nVolSwing = static_cast<uint8>(std::min<uint32>(ins->nVolSwing * 100 / 64, 100));
+			ins->nVolSwing = static_cast<uint8>(std::min(static_cast<uint32>(ins->nVolSwing * 100 / 64), uint32(100)));
 
-			if(!compatModeIT || m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 18, 00, 00))
+			if(!compatModeIT || m_dwLastSavedWithVersion < MPT_V("1.18.00.00"))
 			{
 				// Previously, Pitch/Pan Separation was only half depth (plot twist: it was actually only quarter depth).
 				// This was corrected in compatible mode in OpenMPT 1.18, and in OpenMPT 1.20 it is corrected in normal mode as well.
-				ins->nPPS = (ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2;
+				ins->nPPS = static_cast<int8>((ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2);
 			}
 
-			if(!compatModeIT || m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 17, 03, 02))
+			if(!compatModeIT || m_dwLastSavedWithVersion < MPT_V("1.17.03.02"))
 			{
 				// IT compatibility 24. Short envelope loops
 				// Previously, the pitch / filter envelope loop handling was broken, the loop was shortened by a tick (like in XM).
@@ -272,7 +281,7 @@ void CSoundFile::UpgradeModule()
 				ins->GetEnvelope(ENV_PITCH).Convert(MOD_TYPE_XM, GetType());
 			}
 
-			if(m_dwLastSavedWithVersion >= MAKE_VERSION_NUMERIC(1, 17, 00, 00) && m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 17, 02, 50))
+			if(m_dwLastSavedWithVersion >= MPT_V("1.17.00.00") && m_dwLastSavedWithVersion < MPT_V("1.17.02.50"))
 			{
 				// If there are any plugins that can receive volume commands, enable volume bug emulation.
 				if(ins->nMixPlug && ins->HasValidMIDIChannel())
@@ -281,7 +290,7 @@ void CSoundFile::UpgradeModule()
 				}
 			}
 
-			if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 17, 02, 50) && (ins->nVolSwing | ins->nPanSwing | ins->nCutSwing | ins->nResSwing))
+			if(m_dwLastSavedWithVersion < MPT_V("1.17.02.50") && (ins->nVolSwing | ins->nPanSwing | ins->nCutSwing | ins->nResSwing))
 			{
 				// If there are any instruments with random variation, enable the old random variation behaviour.
 				m_playBehaviour.set(kMPTOldSwingBehaviour);
@@ -289,7 +298,7 @@ void CSoundFile::UpgradeModule()
 			}
 		}
 
-		if((GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT)) && (m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 17, 03, 02) || !compatModeIT))
+		if((GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT)) && (m_dwLastSavedWithVersion < MPT_V("1.17.03.02") || !compatModeIT))
 		{
 			// In the IT format, a sweep value of 0 shouldn't apply vibrato at all. Previously, a value of 0 was treated as "no sweep".
 			// In OpenMPT 1.17.03.02, this was corrected in compatible mode, in OpenMPT 1.20 it is corrected in normal mode as well,
@@ -307,8 +316,8 @@ void CSoundFile::UpgradeModule()
 		m_MidiCfg.UpgradeMacros();
 	}
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 20, 02, 10)
-		&& m_dwLastSavedWithVersion != MAKE_VERSION_NUMERIC(1, 20, 00, 00)
+	if(m_dwLastSavedWithVersion < MPT_V("1.20.02.10")
+		&& m_dwLastSavedWithVersion != MPT_V("1.20.00.00")
 		&& (GetType() & (MOD_TYPE_XM | MOD_TYPE_IT | MOD_TYPE_MPT)))
 	{
 		bool instrPlugs = false;
@@ -327,8 +336,8 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 22, 03, 12)
-		&& m_dwLastSavedWithVersion != MAKE_VERSION_NUMERIC(1, 22, 00, 00)
+	if(m_dwLastSavedWithVersion < MPT_V("1.22.03.12")
+		&& m_dwLastSavedWithVersion != MPT_V("1.22.00.00")
 		&& (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT))
 		&& (m_playBehaviour[MSF_COMPATIBLE_PLAY] || m_playBehaviour[kMPTOldSwingBehaviour]))
 	{
@@ -344,17 +353,30 @@ void CSoundFile::UpgradeModule()
 	}
 
 #ifndef NO_PLUGINS
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 22, 07, 01))
+	bool hasAnyPlugins = false;
+	if(GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT | MOD_TYPE_XM))
+	{
+		for(auto &plugin : m_MixPlugins)
+		{
+			if(plugin.IsValidPlugin())
+			{
+				hasAnyPlugins = true;
+				break;
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.22.07.01") && hasAnyPlugins)
 	{
 		// Convert ANSI plugin path names to UTF-8 (irrelevant in probably 99% of all cases anyway, I think I've never seen a VST plugin with a non-ASCII file name)
-		for(PLUGINDEX i = 0; i < MAX_MIXPLUGINS; i++)
+		for(auto &plugin : m_MixPlugins)
 		{
 #if defined(MODPLUG_TRACKER)
-			const std::string name = mpt::ToCharset(mpt::CharsetUTF8, mpt::CharsetLocale, m_MixPlugins[i].Info.szLibraryName);
+			const std::string name = mpt::ToCharset(mpt::Charset::UTF8, mpt::Charset::Locale, plugin.Info.szLibraryName);
 #else
-			const std::string name = mpt::ToCharset(mpt::CharsetUTF8, mpt::CharsetWindows1252, m_MixPlugins[i].Info.szLibraryName);
+			const std::string name = mpt::ToCharset(mpt::Charset::UTF8, mpt::Charset::Windows1252, plugin.Info.szLibraryName);
 #endif
-			mpt::String::Copy(m_MixPlugins[i].Info.szLibraryName, name);
+			plugin.Info.szLibraryName = name;
 		}
 	}
 #endif // NO_PLUGINS
@@ -363,15 +385,15 @@ void CSoundFile::UpgradeModule()
 	// Starting from OpenMPT 1.23.01.04, FT2-style panning has its own mix mode instead.
 	if(GetType() == MOD_TYPE_XM)
 	{
-		if(m_dwLastSavedWithVersion >= MAKE_VERSION_NUMERIC(1, 22, 07, 19)
-			&& m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 23, 01, 04)
-			&& GetMixLevels() == mixLevelsCompatible)
+		if(m_dwLastSavedWithVersion >= MPT_V("1.22.07.19")
+			&& m_dwLastSavedWithVersion < MPT_V("1.23.01.04")
+			&& GetMixLevels() == MixLevels::Compatible)
 		{
-			SetMixLevels(mixLevelsCompatibleFT2);
+			SetMixLevels(MixLevels::CompatibleFT2);
 		}
 	}
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 25, 00, 07) && m_dwLastSavedWithVersion != MAKE_VERSION_NUMERIC(1, 25, 00, 00))
+	if(m_dwLastSavedWithVersion < MPT_V("1.25.00.07") && m_dwLastSavedWithVersion != MPT_V("1.25.00.00"))
 	{
 		// Instrument plugins can now receive random volume variation.
 		// For old instruments, disable volume swing in case there was no sample associated.
@@ -380,9 +402,9 @@ void CSoundFile::UpgradeModule()
 			if(Instruments[i] != nullptr && Instruments[i]->nVolSwing != 0 && Instruments[i]->nMidiChannel != MidiNoChannel)
 			{
 				bool hasSample = false;
-				for(size_t k = 0; k < CountOf(Instruments[k]->Keyboard); k++)
+				for(auto smp : Instruments[i]->Keyboard)
 				{
-					if(Instruments[i]->Keyboard[k] != 0)
+					if(smp != 0)
 					{
 						hasSample = true;
 						break;
@@ -396,19 +418,57 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 26, 00, 00))
+	if(m_dwLastSavedWithVersion < MPT_V("1.26.00.00"))
 	{
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if(Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
 			ModInstrument *ins = Instruments[i];
+			if(!ins)
+				continue;
 			// Even after fixing it in OpenMPT 1.18, instrument PPS was only half the depth.
-			ins->nPPS = (ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2;
+			ins->nPPS = static_cast<int8>((ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2);
 
 			// OpenMPT 1.18 fixed the depth of random pan in compatible mode.
 			// OpenMPT 1.26 fixes it in normal mode too.
-			if(!compatModeIT || m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 18, 00, 00))
+			if(!compatModeIT || m_dwLastSavedWithVersion < MPT_V("1.18.00.00"))
 			{
-				ins->nPanSwing = (ins->nPanSwing + 3) / 4u;
+				ins->nPanSwing = static_cast<uint8>((ins->nPanSwing + 3) / 4u);
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.28.00.12"))
+	{
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
+		{
+			if(Instruments[i] != nullptr && Instruments[i]->VolEnv.nReleaseNode != ENV_RELEASE_NODE_UNSET)
+			{
+				m_playBehaviour.set(kLegacyReleaseNode);
+				break;
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.28.03.04"))
+	{
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
+		{
+			if(Instruments[i] != nullptr && (Instruments[i]->pluginVolumeHandling == PLUGIN_VOLUMEHANDLING_MIDI || Instruments[i]->pluginVolumeHandling == PLUGIN_VOLUMEHANDLING_DRYWET))
+			{
+				m_playBehaviour.set(kMIDIVolumeOnNoteOffBug);
+				break;
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.30.00.54"))
+	{
+		for(SAMPLEINDEX i = 1; i <= GetNumSamples(); i++)
+		{
+			if(Samples[i].HasSampleData() && Samples[i].uFlags[CHN_PINGPONGLOOP | CHN_PINGPONGSUSTAIN])
+			{
+				m_playBehaviour.set(kImprecisePingPongLoops);
+				break;
 			}
 		}
 	}
@@ -422,98 +482,98 @@ void CSoundFile::UpgradeModule()
 	struct PlayBehaviourVersion
 	{
 		PlayBehaviour behaviour;
-		MptVersion::VersionNum version;
+		Version version;
 	};
 	
-	if(compatModeIT && m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 26, 00, 00))
+	if(compatModeIT && m_dwLastSavedWithVersion < MPT_V("1.26.00.00"))
 	{
 		// Pre-1.26: Detailed compatibility flags did not exist.
 		static constexpr PlayBehaviourVersion behaviours[] =
 		{
-			{ kTempoClamp,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kPerChannelGlobalVolSlide,		MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kPanOverride,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITInstrWithoutNote,				MAKE_VERSION_NUMERIC(1, 17, 02, 46) },
-			{ kITVolColFinePortamento,			MAKE_VERSION_NUMERIC(1, 17, 02, 49) },
-			{ kITArpeggio,						MAKE_VERSION_NUMERIC(1, 17, 02, 49) },
-			{ kITOutOfRangeDelay,				MAKE_VERSION_NUMERIC(1, 17, 02, 49) },
-			{ kITPortaMemoryShare,				MAKE_VERSION_NUMERIC(1, 17, 02, 49) },
-			{ kITPatternLoopTargetReset,		MAKE_VERSION_NUMERIC(1, 17, 02, 49) },
-			{ kITFT2PatternLoop,				MAKE_VERSION_NUMERIC(1, 17, 02, 49) },
-			{ kITPingPongNoReset,				MAKE_VERSION_NUMERIC(1, 17, 02, 51) },
-			{ kITEnvelopeReset,					MAKE_VERSION_NUMERIC(1, 17, 02, 51) },
-			{ kITClearOldNoteAfterCut,			MAKE_VERSION_NUMERIC(1, 17, 02, 52) },
-			{ kITVibratoTremoloPanbrello,		MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITTremor,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITRetrigger,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITMultiSampleBehaviour,			MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITPortaTargetReached,			MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITPatternLoopBreak,				MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITOffset,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITSwingBehaviour,				MAKE_VERSION_NUMERIC(1, 18, 00, 00) },
-			{ kITNNAReset,						MAKE_VERSION_NUMERIC(1, 18, 00, 00) },
-			{ kITSCxStopsSample,				MAKE_VERSION_NUMERIC(1, 18, 00, 01) },
-			{ kITEnvelopePositionHandling,		MAKE_VERSION_NUMERIC(1, 18, 01, 00) },
-			{ kITPortamentoInstrument,			MAKE_VERSION_NUMERIC(1, 19, 00, 01) },
-			{ kITPingPongMode,					MAKE_VERSION_NUMERIC(1, 19, 00, 21) },
-			{ kITRealNoteMapping,				MAKE_VERSION_NUMERIC(1, 19, 00, 30) },
-			{ kITHighOffsetNoRetrig,			MAKE_VERSION_NUMERIC(1, 20, 00, 14) },
-			{ kITFilterBehaviour,				MAKE_VERSION_NUMERIC(1, 20, 00, 35) },
-			{ kITNoSurroundPan,					MAKE_VERSION_NUMERIC(1, 20, 00, 53) },
-			{ kITShortSampleRetrig,				MAKE_VERSION_NUMERIC(1, 20, 00, 54) },
-			{ kITPortaNoNote,					MAKE_VERSION_NUMERIC(1, 20, 00, 56) },
-			{ kRowDelayWithNoteDelay,			MAKE_VERSION_NUMERIC(1, 20, 00, 76) },
-			{ kITDontResetNoteOffOnPorta,		MAKE_VERSION_NUMERIC(1, 20, 02, 06) },
-			{ kITVolColMemory,					MAKE_VERSION_NUMERIC(1, 21, 01, 16) },
-			{ kITPortamentoSwapResetsPos,		MAKE_VERSION_NUMERIC(1, 21, 01, 25) },
-			{ kITEmptyNoteMapSlot,				MAKE_VERSION_NUMERIC(1, 21, 01, 25) },
-			{ kITFirstTickHandling,				MAKE_VERSION_NUMERIC(1, 22, 07, 09) },
-			{ kITSampleAndHoldPanbrello,		MAKE_VERSION_NUMERIC(1, 22, 07, 19) },
-			{ kITClearPortaTarget,				MAKE_VERSION_NUMERIC(1, 23, 04, 03) },
-			{ kITPanbrelloHold,					MAKE_VERSION_NUMERIC(1, 24, 01, 06) },
-			{ kITPanningReset,					MAKE_VERSION_NUMERIC(1, 24, 01, 06) },
-			{ kITPatternLoopWithJumps,			MAKE_VERSION_NUMERIC(1, 25, 00, 19) },
+			{ kTempoClamp,                    MPT_V("1.17.03.02") },
+			{ kPerChannelGlobalVolSlide,      MPT_V("1.17.03.02") },
+			{ kPanOverride,                   MPT_V("1.17.03.02") },
+			{ kITInstrWithoutNote,            MPT_V("1.17.02.46") },
+			{ kITVolColFinePortamento,        MPT_V("1.17.02.49") },
+			{ kITArpeggio,                    MPT_V("1.17.02.49") },
+			{ kITOutOfRangeDelay,             MPT_V("1.17.02.49") },
+			{ kITPortaMemoryShare,            MPT_V("1.17.02.49") },
+			{ kITPatternLoopTargetReset,      MPT_V("1.17.02.49") },
+			{ kITFT2PatternLoop,              MPT_V("1.17.02.49") },
+			{ kITPingPongNoReset,             MPT_V("1.17.02.51") },
+			{ kITEnvelopeReset,               MPT_V("1.17.02.51") },
+			{ kITClearOldNoteAfterCut,        MPT_V("1.17.02.52") },
+			{ kITVibratoTremoloPanbrello,     MPT_V("1.17.03.02") },
+			{ kITTremor,                      MPT_V("1.17.03.02") },
+			{ kITRetrigger,                   MPT_V("1.17.03.02") },
+			{ kITMultiSampleBehaviour,        MPT_V("1.17.03.02") },
+			{ kITPortaTargetReached,          MPT_V("1.17.03.02") },
+			{ kITPatternLoopBreak,            MPT_V("1.17.03.02") },
+			{ kITOffset,                      MPT_V("1.17.03.02") },
+			{ kITSwingBehaviour,              MPT_V("1.18.00.00") },
+			{ kITNNAReset,                    MPT_V("1.18.00.00") },
+			{ kITSCxStopsSample,              MPT_V("1.18.00.01") },
+			{ kITEnvelopePositionHandling,    MPT_V("1.18.01.00") },
+			{ kITPortamentoInstrument,        MPT_V("1.19.00.01") },
+			{ kITPingPongMode,                MPT_V("1.19.00.21") },
+			{ kITRealNoteMapping,             MPT_V("1.19.00.30") },
+			{ kITHighOffsetNoRetrig,          MPT_V("1.20.00.14") },
+			{ kITFilterBehaviour,             MPT_V("1.20.00.35") },
+			{ kITNoSurroundPan,               MPT_V("1.20.00.53") },
+			{ kITShortSampleRetrig,           MPT_V("1.20.00.54") },
+			{ kITPortaNoNote,                 MPT_V("1.20.00.56") },
+			{ kRowDelayWithNoteDelay,         MPT_V("1.20.00.76") },
+			{ kITFT2DontResetNoteOffOnPorta,  MPT_V("1.20.02.06") },
+			{ kITVolColMemory,                MPT_V("1.21.01.16") },
+			{ kITPortamentoSwapResetsPos,     MPT_V("1.21.01.25") },
+			{ kITEmptyNoteMapSlot,            MPT_V("1.21.01.25") },
+			{ kITFirstTickHandling,           MPT_V("1.22.07.09") },
+			{ kITSampleAndHoldPanbrello,      MPT_V("1.22.07.19") },
+			{ kITClearPortaTarget,            MPT_V("1.23.04.03") },
+			{ kITPanbrelloHold,               MPT_V("1.24.01.06") },
+			{ kITPanningReset,                MPT_V("1.24.01.06") },
+			{ kITPatternLoopWithJumpsOld,     MPT_V("1.25.00.19") },
 		};
 
 		for(const auto &b : behaviours)
 		{
-			m_playBehaviour.set(b.behaviour, (m_dwLastSavedWithVersion >= b.version || m_dwLastSavedWithVersion == (b.version & 0xFFFF0000)));
+			m_playBehaviour.set(b.behaviour, (m_dwLastSavedWithVersion >= b.version || m_dwLastSavedWithVersion == b.version.Masked(0xFFFF0000u)));
 		}
-	} else if(compatModeXM && m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 26, 00, 00))
+	} else if(compatModeXM && m_dwLastSavedWithVersion < MPT_V("1.26.00.00"))
 	{
 		// Pre-1.26: Detailed compatibility flags did not exist.
 		static constexpr PlayBehaviourVersion behaviours[] =
 		{
-			{ kTempoClamp,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kPerChannelGlobalVolSlide,		MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kPanOverride,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kITFT2PatternLoop,				MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2Arpeggio,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2Retrigger,					MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2VolColVibrato,				MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2PortaNoNote,					MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2KeyOff,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2PanSlide,						MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2OffsetOutOfRange,				MAKE_VERSION_NUMERIC(1, 17, 03, 02) },
-			{ kFT2RestrictXCommand,				MAKE_VERSION_NUMERIC(1, 18, 00, 00) },
-			{ kFT2RetrigWithNoteDelay,			MAKE_VERSION_NUMERIC(1, 18, 00, 00) },
-			{ kFT2SetPanEnvPos,					MAKE_VERSION_NUMERIC(1, 18, 00, 00) },
-			{ kFT2PortaIgnoreInstr,				MAKE_VERSION_NUMERIC(1, 18, 00, 01) },
-			{ kFT2VolColMemory,					MAKE_VERSION_NUMERIC(1, 18, 01, 00) },
-			{ kFT2LoopE60Restart,				MAKE_VERSION_NUMERIC(1, 18, 02, 01) },
-			{ kFT2ProcessSilentChannels,		MAKE_VERSION_NUMERIC(1, 18, 02, 01) },
-			{ kFT2ReloadSampleSettings,			MAKE_VERSION_NUMERIC(1, 20, 00, 36) },
-			{ kFT2PortaDelay,					MAKE_VERSION_NUMERIC(1, 20, 00, 40) },
-			{ kFT2Transpose,					MAKE_VERSION_NUMERIC(1, 20, 00, 62) },
-			{ kFT2PatternLoopWithJumps,			MAKE_VERSION_NUMERIC(1, 20, 00, 69) },
-			{ kFT2PortaTargetNoReset,			MAKE_VERSION_NUMERIC(1, 20, 00, 69) },
-			{ kFT2EnvelopeEscape,				MAKE_VERSION_NUMERIC(1, 20, 00, 77) },
-			{ kFT2Tremor,						MAKE_VERSION_NUMERIC(1, 20, 01, 11) },
-			{ kFT2OutOfRangeDelay,				MAKE_VERSION_NUMERIC(1, 20, 02, 02) },
-			{ kFT2Periods,						MAKE_VERSION_NUMERIC(1, 22, 03, 01) },
-			{ kFT2PanWithDelayedNoteOff,		MAKE_VERSION_NUMERIC(1, 22, 03, 02) },
-			{ kFT2VolColDelay,					MAKE_VERSION_NUMERIC(1, 22, 07, 19) },
-			{ kFT2FinetunePrecision,			MAKE_VERSION_NUMERIC(1, 22, 07, 19) },
+			{ kTempoClamp,               MPT_V("1.17.03.02") },
+			{ kPerChannelGlobalVolSlide, MPT_V("1.17.03.02") },
+			{ kPanOverride,              MPT_V("1.17.03.02") },
+			{ kITFT2PatternLoop,         MPT_V("1.17.03.02") },
+			{ kFT2Arpeggio,              MPT_V("1.17.03.02") },
+			{ kFT2Retrigger,             MPT_V("1.17.03.02") },
+			{ kFT2VolColVibrato,         MPT_V("1.17.03.02") },
+			{ kFT2PortaNoNote,           MPT_V("1.17.03.02") },
+			{ kFT2KeyOff,                MPT_V("1.17.03.02") },
+			{ kFT2PanSlide,              MPT_V("1.17.03.02") },
+			{ kFT2ST3OffsetOutOfRange,   MPT_V("1.17.03.02") },
+			{ kFT2RestrictXCommand,      MPT_V("1.18.00.00") },
+			{ kFT2RetrigWithNoteDelay,   MPT_V("1.18.00.00") },
+			{ kFT2SetPanEnvPos,          MPT_V("1.18.00.00") },
+			{ kFT2PortaIgnoreInstr,      MPT_V("1.18.00.01") },
+			{ kFT2VolColMemory,          MPT_V("1.18.01.00") },
+			{ kFT2LoopE60Restart,        MPT_V("1.18.02.01") },
+			{ kFT2ProcessSilentChannels, MPT_V("1.18.02.01") },
+			{ kFT2ReloadSampleSettings,  MPT_V("1.20.00.36") },
+			{ kFT2PortaDelay,            MPT_V("1.20.00.40") },
+			{ kFT2Transpose,             MPT_V("1.20.00.62") },
+			{ kFT2PatternLoopWithJumps,  MPT_V("1.20.00.69") },
+			{ kFT2PortaTargetNoReset,    MPT_V("1.20.00.69") },
+			{ kFT2EnvelopeEscape,        MPT_V("1.20.00.77") },
+			{ kFT2Tremor,                MPT_V("1.20.01.11") },
+			{ kFT2OutOfRangeDelay,       MPT_V("1.20.02.02") },
+			{ kFT2Periods,               MPT_V("1.22.03.01") },
+			{ kFT2PanWithDelayedNoteOff, MPT_V("1.22.03.02") },
+			{ kFT2VolColDelay,           MPT_V("1.22.07.19") },
+			{ kFT2FinetunePrecision,     MPT_V("1.22.07.19") },
 		};
 
 		for(const auto &b : behaviours)
@@ -527,16 +587,23 @@ void CSoundFile::UpgradeModule()
 		// The following behaviours were added in/after OpenMPT 1.26, so are not affected by the upgrade mechanism above.
 		static constexpr PlayBehaviourVersion behaviours[] =
 		{
-			{ kITInstrWithNoteOff,				MAKE_VERSION_NUMERIC(1, 26, 00, 01) },
-			{ kITMultiSampleInstrumentNumber,	MAKE_VERSION_NUMERIC(1, 27, 00, 27) },
+			{ kITInstrWithNoteOff,            MPT_V("1.26.00.01") },
+			{ kITMultiSampleInstrumentNumber, MPT_V("1.27.00.27") },
+			{ kITInstrWithNoteOffOldEffects,  MPT_V("1.28.02.06") },
+			{ kITDoNotOverrideChannelPan,     MPT_V("1.29.00.22") },
+			{ kITPatternLoopWithJumps,        MPT_V("1.29.00.32") },
+			{ kITDCTBehaviour,                MPT_V("1.29.00.57") },
+			{ kITPitchPanSeparation,          MPT_V("1.30.00.53") },
+			{ kITResetFilterOnPortaSmpChange, MPT_V("1.30.08.02") },
+			{ kITInitialNoteMemory,           MPT_V("1.31.00.25") },
 		};
 
 		for(const auto &b : behaviours)
 		{
-			if(m_dwLastSavedWithVersion < (b.version & 0xFFFF0000))
+			if(m_dwLastSavedWithVersion < b.version.Masked(0xFFFF0000u))
 				m_playBehaviour.reset(b.behaviour);
 			// Full version information available, i.e. not compatibility-exported.
-			else if(m_dwLastSavedWithVersion > (b.version & 0xFFFF0000) && m_dwLastSavedWithVersion < b.version)
+			else if(m_dwLastSavedWithVersion > b.version.Masked(0xFFFF0000u) && m_dwLastSavedWithVersion < b.version)
 				m_playBehaviour.reset(b.behaviour);
 		}
 	} else if(GetType() == MOD_TYPE_XM)
@@ -544,10 +611,14 @@ void CSoundFile::UpgradeModule()
 		// The following behaviours were added after OpenMPT 1.26, so are not affected by the upgrade mechanism above.
 		static constexpr PlayBehaviourVersion behaviours[] =
 		{
-			{ kFT2NoteOffFlags,					MAKE_VERSION_NUMERIC(1, 27, 00, 27) },
-			{ kRowDelayWithNoteDelay,			MAKE_VERSION_NUMERIC(1, 27, 00, 37) },
-			{ kFT2TremoloRampWaveform,			MAKE_VERSION_NUMERIC(1, 27, 00, 37) },
-			{ kFT2PortaUpDownMemory,			MAKE_VERSION_NUMERIC(1, 27, 00, 37) },
+			{ kFT2NoteOffFlags,              MPT_V("1.27.00.27") },
+			{ kRowDelayWithNoteDelay,        MPT_V("1.27.00.37") },
+			{ kFT2MODTremoloRampWaveform,    MPT_V("1.27.00.37") },
+			{ kFT2PortaUpDownMemory,         MPT_V("1.27.00.37") },
+			{ kFT2PanSustainRelease,         MPT_V("1.28.00.09") },
+			{ kFT2NoteDelayWithoutInstr,     MPT_V("1.28.00.44") },
+			{ kITFT2DontResetNoteOffOnPorta, MPT_V("1.29.00.34") },
+			{ kFT2PortaResetDirection,       MPT_V("1.30.00.40") },
 		};
 
 		for(const auto &b : behaviours)
@@ -560,13 +631,18 @@ void CSoundFile::UpgradeModule()
 		// We do not store any of these flags in S3M files.
 		static constexpr PlayBehaviourVersion behaviours[] =
 		{
-			{ kST3NoMutedChannels,		MAKE_VERSION_NUMERIC(1, 18, 00, 00) },
-			{ kST3EffectMemory,			MAKE_VERSION_NUMERIC(1, 20, 00, 00) },
-			{ kRowDelayWithNoteDelay,	MAKE_VERSION_NUMERIC(1, 20, 00, 00) },
-			{ kST3PortaSampleChange,	MAKE_VERSION_NUMERIC(1, 22, 00, 00) },
-			{ kST3VibratoMemory,		MAKE_VERSION_NUMERIC(1, 26, 00, 00) },
-			{ kITPanbrelloHold,			MAKE_VERSION_NUMERIC(1, 26, 00, 00) },
-			{ KST3PortaAfterArpeggio,	MAKE_VERSION_NUMERIC(1, 27, 00, 00) },
+			{ kST3NoMutedChannels,         MPT_V("1.18.00.00") },
+			{ kST3EffectMemory,            MPT_V("1.20.00.00") },
+			{ kRowDelayWithNoteDelay,      MPT_V("1.20.00.00") },
+			{ kST3PortaSampleChange,       MPT_V("1.22.00.00") },
+			{ kST3VibratoMemory,           MPT_V("1.26.00.00") },
+			{ kITPanbrelloHold,            MPT_V("1.26.00.00") },
+			{ KST3PortaAfterArpeggio,      MPT_V("1.27.00.00") },
+			{ kST3OffsetWithoutInstrument, MPT_V("1.28.00.00") },
+			{ kST3RetrigAfterNoteCut,      MPT_V("1.29.00.00") },
+			{ kFT2ST3OffsetOutOfRange,     MPT_V("1.29.00.00") },
+			{ kApplyUpperPeriodLimit,      MPT_V("1.30.00.45") },
+			{ kST3TonePortaWithAdlibNote,  MPT_V("1.31.00.13") },
 		};
 
 		for(const auto &b : behaviours)
@@ -576,7 +652,13 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
-	if(m_dwLastSavedWithVersion >= MAKE_VERSION_NUMERIC(1, 27, 00, 27) && m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 27, 00, 49))
+	if(GetType() == MOD_TYPE_XM && m_dwLastSavedWithVersion < MPT_V("1.19.00.00"))
+	{
+		// This bug was introduced sometime between 1.18.03.00 and 1.19.01.00
+		m_playBehaviour.set(kFT2NoteDelayWithoutInstr);
+	}
+
+	if(m_dwLastSavedWithVersion >= MPT_V("1.27.00.27") && m_dwLastSavedWithVersion < MPT_V("1.27.00.49"))
 	{
 		// OpenMPT 1.27 inserted some IT/FT2 flags before the S3M flags that are never saved to files anyway, to keep the flag IDs a bit more compact.
 		// However, it was overlooked that these flags would still be read by OpenMPT 1.26 and thus S3M-specific behaviour would be enabled in IT/XM files.
@@ -589,27 +671,105 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 17, 00, 00))
+	if(m_dwLastSavedWithVersion < MPT_V("1.17.00.00"))
 	{
 		// MPT 1.16 has a maximum tempo of 255.
 		m_playBehaviour.set(kTempoClamp);
-	} else if(m_dwLastSavedWithVersion >= MAKE_VERSION_NUMERIC(1, 17, 00, 00) && m_dwLastSavedWithVersion <= MAKE_VERSION_NUMERIC(1, 20, 01, 03) && m_dwLastSavedWithVersion != MAKE_VERSION_NUMERIC(1, 20, 00, 00))
+	} else if(m_dwLastSavedWithVersion >= MPT_V("1.17.00.00") && m_dwLastSavedWithVersion <= MPT_V("1.20.01.03") && m_dwLastSavedWithVersion != MPT_V("1.20.00.00"))
 	{
 		// OpenMPT introduced some "fixes" that execute regular portamentos also at speed 1.
 		m_playBehaviour.set(kSlidesAtSpeed1);
 	}
 
-	if(m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 24, 00, 00))
+	if(m_SongFlags[SONG_LINEARSLIDES])
 	{
-		// No frequency slides in Hz before OpenMPT 1.24
-		m_playBehaviour.reset(kHertzInLinearMode);
-	} else if(m_dwLastSavedWithVersion >= MAKE_VERSION_NUMERIC(1, 24, 00, 00) && m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 26, 00, 00) && (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT)))
+		if(m_dwLastSavedWithVersion < MPT_V("1.24.00.00"))
+		{
+			// No frequency slides in Hz before OpenMPT 1.24
+			m_playBehaviour.reset(kPeriodsAreHertz);
+		} else if(m_dwLastSavedWithVersion >= MPT_V("1.24.00.00") && m_dwLastSavedWithVersion < MPT_V("1.26.00.00") && (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT)))
+		{
+			// Frequency slides were always in Hz rather than periods in this version range.
+			m_playBehaviour.set(kPeriodsAreHertz);
+		}
+	} else
 	{
-		// Frequency slides were always in Hz rather than periods in this version range.
-		m_playBehaviour.set(kHertzInLinearMode);
+		if(m_dwLastSavedWithVersion < MPT_V("1.30.00.36") && m_dwLastSavedWithVersion != MPT_V("1.30.00.00"))
+		{
+			// No frequency slides in Hz before OpenMPT 1.30
+			m_playBehaviour.reset(kPeriodsAreHertz);
+		}
 	}
 
-	if(m_dwLastSavedWithVersion >= MAKE_VERSION_NUMERIC(1, 27, 00, 00) && m_dwLastSavedWithVersion < MAKE_VERSION_NUMERIC(1, 30, 06, 00) && (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT | MOD_TYPE_XM)))
+	if(m_playBehaviour[kITEnvelopePositionHandling]
+		&& m_dwLastSavedWithVersion >= MPT_V("1.23.01.02") && m_dwLastSavedWithVersion < MPT_V("1.28.00.43"))
+	{
+		// Bug that effectively clamped the release node to the sustain end
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
+		{
+			if(Instruments[i] != nullptr
+				&& Instruments[i]->VolEnv.nReleaseNode != ENV_RELEASE_NODE_UNSET
+				&& Instruments[i]->VolEnv.dwFlags[ENV_SUSTAIN]
+				&& Instruments[i]->VolEnv.nReleaseNode > Instruments[i]->VolEnv.nSustainEnd)
+			{
+				m_playBehaviour.set(kReleaseNodePastSustainBug);
+				break;
+			}
+		}
+	}
+
+	if(GetType() & (MOD_TYPE_MPT | MOD_TYPE_S3M))
+	{
+		for(SAMPLEINDEX i = 1; i <= GetNumSamples(); i++)
+		{
+			if(Samples[i].uFlags[CHN_ADLIB])
+			{
+				if(GetType() == MOD_TYPE_MPT && GetNumInstruments() && m_dwLastSavedWithVersion >= MPT_V("1.28.00.20") && m_dwLastSavedWithVersion <= MPT_V("1.29.00.55"))
+					m_playBehaviour.set(kOPLNoResetAtEnvelopeEnd);
+				if(m_dwLastSavedWithVersion <= MPT_V("1.30.00.34") && m_dwLastSavedWithVersion != MPT_V("1.30"))
+					m_playBehaviour.reset(kOPLNoteOffOnNoteChange);
+				if(GetType() == MOD_TYPE_S3M && m_dwLastSavedWithVersion < MPT_V("1.29"))
+					m_playBehaviour.set(kOPLRealRetrig);
+				else if(GetType() != MOD_TYPE_S3M)
+					m_playBehaviour.reset(kOPLRealRetrig);
+				break;
+			}
+		}
+	}
+
+#ifndef NO_PLUGINS
+	if(m_dwLastSavedWithVersion >= MPT_V("1.27.00.42") && m_dwLastSavedWithVersion < MPT_V("1.30.00.46") && hasAnyPlugins)
+	{
+		// The Flanger DMO plugin is almost identical to the Chorus... but only almost.
+		// The effect implementation was the same in OpenMPT 1.27-1.29, now it isn't anymore.
+		// As the old implementation continues to exist for the Chorus plugin, there is a legacy wrapper for the Flanger plugin.
+		for(auto &plugin : m_MixPlugins)
+		{
+			if(plugin.Info.dwPluginId1 == kDmoMagic && plugin.Info.dwPluginId2 == int32(0xEFCA3D92) && plugin.pluginData.size() == 32)
+				plugin.Info.szLibraryName = "Flanger (Legacy)";
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.30.00.54") && hasAnyPlugins)
+	{
+		// Currently active program and bank is assumed to be 1 when starting playback
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
+		{
+			if(Instruments[i] && (Instruments[i]->nMidiProgram == 1 || Instruments[i]->wMidiBank == 1))
+			{
+				m_playBehaviour.set(kPluginDefaultProgramAndBank1);
+				break;
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.31.00.09") && hasAnyPlugins)
+	{
+		// Old-style plugin tone portamento
+		m_playBehaviour.set(kPluginIgnoreTonePortamento);
+	}
+
+	if(m_dwLastSavedWithVersion >= MPT_V("1.27") && m_dwLastSavedWithVersion < MPT_V("1.30.06.00") && hasAnyPlugins)
 	{
 		// Fix off-by-one delay length in older Echo DMO emulation
 		for(auto &plugin : m_MixPlugins)
@@ -626,7 +786,7 @@ void CSoundFile::UpgradeModule()
 			}
 		}
 	}
+#endif  // NO_PLUGINS
 }
 
-
 OPENMPT_NAMESPACE_END
